/*
 *  hnv_search_tree.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/15/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 */

#include <stdio.h>
#include <stdlib.h>
#include "hnv_main.h"
#include "hnv_computation_main.h"
#include "hnv_tetrahedron.h"


struct search_node_chunk {
	int array_size;
	struct search_node * node_array;
	int chunk_ct;
	struct search_node_chunk * next;
};

struct search_node {
	struct hnv_coor_3d_int locator;
	struct tetrahedron * pt;
	
	struct search_node * left;
	struct search_node * right;
};


//===============================================
// Private data
//===============================================

struct search_node_chunk search_node_pool_head;
struct search_node unused_search_node_head;
struct search_node * search_tree_root;


//===============================================
// Private methods
//===============================================
// 
// void print_search_node_chunk(struct search_node_chunk * s_chunk);
// void print_snode_chunk_list(struct search_node_chunk * chunk);
// void print_seach_node(struct search_node * snode);
// int get_snode_chunk_ct(struct search_node_chunk * s_chunk);
// void initialize_snode_chunk(struct search_node_chunk * new_chunk);
// void put_snode_chunk(struct search_node_chunk * list_head, 
//			struct search_node_chunk * new_chunk);
// void put_search_node(struct search_node * list_head, struct search_node * snode);
// struct search_node * take_search_node(struct search_node * list_head);
// int allocate_search_nodes();
// void initialize_search_node(struct search_node * snode);
// void set_up_empty_snode_list(struct search_node * list_head);
// void fill_search_node_list(struct search_node * list_head);
// struct search_node * assign_search_node(void);
// void recall_search_node(struct search_node * snode);
// HNV_COMPARE_RESULT which_is_greater(struct search_node * current_snode,
//				       struct search_node * new_snode);
// void walk_seach_node_tree(struct search_node * snode);
//-----------------------------------------------

void print_search_node_chunk(struct search_node_chunk * s_chunk)
{
	fprintf(STDOUT, "print_search_node_chunk() : chunk = %x, array_size = %d, chunk_ct = %d, next chunk = %x\n",
		(unsigned int) s_chunk, s_chunk->array_size, s_chunk->chunk_ct, (unsigned int) s_chunk->next);
	return;
}


void print_snode_chunk_list(struct search_node_chunk * chunk)
{
	struct search_node_chunk * next_chunk;
	int c_ct;
	
	next_chunk = chunk;
	c_ct = 0;
	while (next_chunk != NULL) {
		fprintf(STDOUT, "print_snode_chunk_list() : start of %d-th search node chunk\n", c_ct);
		print_search_node_chunk(next_chunk);
		next_chunk = next_chunk->next;
		c_ct++;
	}
	return;
}


int get_snode_chunk_ct(struct search_node_chunk * s_chunk)
{
	return s_chunk->chunk_ct;
}


void print_seach_node(struct search_node * snode)
{
	fprintf(STDOUT, "print_seach_node() : locator = { %d, %d, %d }, snode = %x\n",
		snode->locator.a, snode->locator.b, snode->locator.c, (unsigned int) snode);
	fprintf(STDOUT, "print_seach_node() : snode->left = %x, snode->right = %x\n",
		(unsigned int) snode->left, (unsigned int) snode->right);
	return;
}


void initialize_snode_chunk(struct search_node_chunk * new_chunk)
{
	new_chunk->array_size = 0;
	new_chunk->node_array = NULL;
	new_chunk->chunk_ct = 0;
	new_chunk->next = NULL;
	return;
}


void put_snode_chunk(struct search_node_chunk * list_head, 
		     struct search_node_chunk * new_chunk)
{
	new_chunk->next = list_head->next;
	list_head->next = new_chunk;
	list_head->chunk_ct ++;
	return;
}


void put_search_node(struct search_node * list_head,
		     struct search_node * snode)
{
	snode->right = list_head->right;
	list_head->right = snode;
	return;
}


struct search_node * take_search_node(struct search_node * list_head)
{
	struct search_node * snode;
	
	snode = list_head->right;
	if (snode == NULL)
		return NULL;
	list_head->right = snode->right;
	snode->right = NULL;
	return snode;
}


void initialize_search_node(struct search_node * snode)
{
	snode->locator.a = 0;
	snode->locator.b = 0;
	snode->locator.c = 0;
	snode->pt = NULL;
	snode->left = NULL;
	snode->right = NULL;
	return;
}


//
// search_node_pool_head -> chunk0
//                            |
//                           node_array[initial_tetra_chunk_size]
//
int allocate_search_nodes(void)
{
	struct search_node_chunk * new_chunk;
	struct search_node * new_node_array;
	int n_ct;

	if (get_param_safe_limit(&my_param) < get_param_init_size(&my_param) + 
	      get_snode_chunk_ct(&search_node_pool_head) * get_param_incre_size(&my_param)) {
		fprintf(STDERR, "allocate_search_nodes() : ERROR safe limit (=%d) over\n", 
			get_param_safe_limit(&my_param));
		return FAIL;
	}
	new_chunk = (struct search_node_chunk *) malloc(sizeof(struct search_node_chunk));
	if (new_chunk == NULL) {
		fprintf(STDERR, 
			"allocate_search_nodes() : ERROR allocation failed (new_chunk)\n");
		return FAIL;
	}
	initialize_snode_chunk(new_chunk);
	if (search_node_pool_head.next == NULL) {
		new_node_array = (struct search_node *) malloc(get_param_init_size(&my_param) * sizeof(struct search_node));
		if (new_node_array == NULL) {
			fprintf(STDERR, 
				"allocate_search_node() : ERROR initial allocation failed (new_node_array[%d])\n",
				get_param_init_size(&my_param));
			return FAIL;
		}
		new_chunk->array_size = get_param_init_size(&my_param);
	} else {
		new_node_array = (struct search_node *) malloc(get_param_incre_size(&my_param) * sizeof(struct search_node));
		if (new_node_array == NULL) {
			fprintf(STDERR, 
				"allocate_search_node() : ERROR incremental allocation failed (new_node_array[%d])\n",
				get_param_incre_size(&my_param));
			return FAIL;
		}
		new_chunk->array_size = get_param_incre_size(&my_param);
	}
	for (n_ct = 0; n_ct < new_chunk->array_size; n_ct++)
		initialize_search_node(new_node_array + n_ct);
	new_chunk->node_array = new_node_array;
	put_snode_chunk(&search_node_pool_head, new_chunk);
	if (0) {
		fprintf(STDOUT, "allocate_search_node() : DEBUG START of search_node_pool_head\n");
		print_snode_chunk_list(&search_node_pool_head);
		fprintf(STDOUT, "allocate_search_node() : DEBUG END of search_node_pool_head\n");
		fprintf(STDOUT, "\n");
	}
	return SUCCESS;
}


//
// list_head -> NULL
//
void set_up_empty_snode_list(struct search_node * list_head)
{
	initialize_search_node(list_head);
	return;
}


//
// list_head -> ... -> (top_chunk->s_node1) -> (top_chunk->s_node0) -> old_snodes
//
void fill_search_node_list(struct search_node * list_head)
{
	struct search_node_chunk * top_chunk;
	int i;
	
	top_chunk = search_node_pool_head.next;
	for (i=0; i < top_chunk->array_size ; i++)
		put_search_node(list_head, top_chunk->node_array + i);
	return;
}


struct search_node * assign_search_node(void)
{
	struct search_node * snode;
	
	snode = take_search_node(&unused_search_node_head);
	if (snode == NULL) {
		if (DEBUG) fprintf(STDOUT, "assign_search_node() : DEBUG allocate new chunk\n");
		if (allocate_search_nodes() == FAIL) {
			fprintf(STDERR, "assign_search_node() : ERROR allocation failed\n");
			exit(1);
		}
		fill_search_node_list(&unused_search_node_head);
		snode = take_search_node(&unused_search_node_head);
		if (snode == NULL) {
			fprintf(STDERR, "assign_search_node() : ERROR snode == NULL\n");
			if (DEBUG) print_snode_chunk_list(&search_node_pool_head);
			exit(1);
		}
	}
	return snode;
}


void recall_search_node(struct search_node * snode)
{
	initialize_search_node(snode);
	put_search_node(&unused_search_node_head, snode);
	return;
}


//
//  {a0, b0, c0} > {a1, b1, c1}
//  <=> a0 > a1,
//      or (a0 == a1) && (b0 > b1)
//      or (a0 == a1) && (b0 == b1) && (c0 > c1)
//
HNV_COMPARE_RESULT which_is_greater(struct search_node * current_snode,
				    struct search_node * new_snode)
{
	if (current_snode->locator.a > new_snode->locator.a) {
		return HNV_LEFT;
	} else if (current_snode->locator.a == new_snode->locator.a) {
		if (current_snode->locator.b > new_snode->locator.b) {
			return HNV_LEFT;
		} else if (current_snode->locator.b == new_snode->locator.b) {
			if (current_snode->locator.c > new_snode->locator.c) {
				return HNV_LEFT;
			} else if (current_snode->locator.c == new_snode->locator.c) {
				return HNV_EQUAL;
			} else {
				return HNV_RIGHT;
			}
		} else {
			return HNV_RIGHT;
		}
	} else {
		return HNV_RIGHT;
	}
}


void walk_seach_node_tree(struct search_node * snode)
{
	if (snode != NULL) {
		walk_seach_node_tree(snode->left);
		print_seach_node(snode);
		walk_seach_node_tree(snode->right);
	}
	return;
}


//===============================================
// Public methods
//===============================================

void print_seach_node_tree(void)
{
	fprintf(STDOUT, "print_seach_node_tree() : START\n");
	walk_seach_node_tree(search_tree_root);
	fprintf(STDOUT, "print_seach_node_tree() : END\n");
	fprintf(STDOUT, "\n");
	return;
}


void initialize_search_node_tree(void)
{
	initialize_snode_chunk(&search_node_pool_head);
	if (allocate_search_nodes() == FAIL) {
		fprintf(STDERR, "initialize_search_node_pool() : ERROR allocation failed\n");
		exit(1);
	}
	set_up_empty_snode_list(&unused_search_node_head);
	fill_search_node_list(&unused_search_node_head);
	if (0) {
		fprintf(STDOUT, "initialize_search_node_tree() : DEBUG START of search_node_pool_head\n");
		if (DEBUG) print_snode_chunk_list(&search_node_pool_head);
		fprintf(STDOUT, "initialize_search_node_tree() : DEBUG END of search_node_pool_head\n");
		fprintf(STDOUT, "\n");
	}
	search_tree_root = NULL;
	return;
}


void finalize_search_node_tree(void)
{
	struct search_node_chunk * next_chunk;
	struct search_node_chunk * save_chunk;
	
	next_chunk = search_node_pool_head.next;
	while (next_chunk != NULL) {
		save_chunk = next_chunk;
		next_chunk = next_chunk->next;
		free(save_chunk);
	}
	search_tree_root = NULL;
	return;
}


//
//      search_tree_root =
//                ------node0------
//                |               |
//            --node1--        --node2--
//            |       |        |       |
//         node11   node12    node21  node22
//
void register_in_search_tree(struct tetrahedron * ft)
{
	struct search_node * new_snode;
	struct search_node * current_snode;
	struct search_node * old_snode;
	HNV_COMPARE_RESULT comp_flag;
	
	new_snode = assign_search_node();
	new_snode->pt = ft;
	get_tetra_locator(&new_snode->locator, ft);
	if (search_tree_root != NULL) {
		current_snode = search_tree_root;
		while (current_snode != NULL) {
			old_snode = current_snode;
			comp_flag = which_is_greater(current_snode, new_snode);
			if (comp_flag == HNV_LEFT) {	// current_snode > new_snode
				current_snode = current_snode->left;
			} else if (comp_flag == HNV_RIGHT) {
				current_snode = current_snode->right;
			} else {
				fprintf(STDERR,
		"register_in_search_tree() : ERROR tetrahedron {%d, %d, %d} is already registered\n",
					new_snode->locator.a, new_snode->locator.b, new_snode->locator.c);
				exit(1);
			}
		}
		if (comp_flag == HNV_LEFT) {	// old_snode > new_snode
			old_snode->left = new_snode;
		} else {
			old_snode->right = new_snode;
		}
	} else {
		search_tree_root = new_snode;
	}
	return;
}


struct tetrahedron * check_if_processed(struct tetrahedron * ft)
{
	struct search_node * new_snode;
	struct search_node * current_snode;
	HNV_COMPARE_RESULT comp_flag;

	new_snode = assign_search_node();
	get_tetra_locator(&new_snode->locator, ft);
	current_snode = search_tree_root;
	while (current_snode != NULL) {
		comp_flag = which_is_greater(current_snode, new_snode);
		if (comp_flag == HNV_LEFT) {	// current_snode > new_snode
			current_snode = current_snode->left;
		} else if (comp_flag == HNV_RIGHT) {
			current_snode = current_snode->right;
		} else {
			recall_search_node(new_snode);
			return current_snode->pt;	// processed
		}
	}
	recall_search_node(new_snode);
	return (struct tetrahedron *) NULL;	// not processed
}

