/*
 *  hnv_tetrahedron.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/15/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 */

#include <stdio.h>
#include <stdlib.h>
#include "hnv_main.h"
#include "hnv_computation_main.h"


struct tetrahedron_chunk {
	int array_size;
	struct tetrahedron * tetra_array;
	
	int chunk_ct;			 // (for list_head)
	struct tetrahedron_chunk * next;
};

struct tetrahedron {
        struct hnv_coor_3d vertices[4];	// {a0, b0, c0}, {a1, b1, c1}, {a2, b2, c2}, {a3, b3, c3}
	int vertex_order[4];	// top-second-third-bottom
	struct hnv_coor_3d_int locator;	// pos of center * 4
	float height;
	HNV_GRADIENT gradient;		// xyz, xyw, xzw, yzw
	HNV_TETRA_CLASS belongs_to;	// class: IN, BD, OUT, UNKNOWN (unused)
	int direction;			// HNV_UP (= +1), HNV_DOWN (= -1), or 0
	int nearby_peak;
	
	int tetra_ct;			// count of tetrahedrons (for list_head)
        struct tetrahedron * next;	// next tetra on the trajectory or bd tetra list
};

//===============================================
// Private data
//===============================================

struct tetrahedron_chunk tetrahedron_pool_head;
struct tetrahedron unused_terahedron_head;


//===============================================
// Public data
//===============================================

struct tetrahedron bd_tetra_list_head;


//===============================================
// Private methods
//===============================================
// 
// void initialize_tetra_chunk(struct tetrahedron_chunk * new_chunk);
// void put_tetra_chunk(struct tetrahedron_chunk * list_head, struct tetrahedron_chunk * new_chunk);
// void initialize_tetrahedron(struct tetrahedron * tetra);
// int allocate_tetrahedrons(void);
// void copy_vertex(struct hnv_coor_3d * tgt_vertex, struct hnv_coor_3d * src_vertex);
// void print_tetra_chunk(struct tetrahedron_chunk * t_chunk);
// void print_tetra_chunk_list(struct tetrahedron_chunk * chunk);
//-----------------------------------------------

void initialize_tetra_chunk(struct tetrahedron_chunk * new_chunk)
{
	new_chunk->array_size = 0;
	new_chunk->tetra_array = NULL;
	new_chunk->chunk_ct = 0;
	new_chunk->next = NULL;
	return;
}


void put_tetra_chunk(struct tetrahedron_chunk * list_head, 
		     struct tetrahedron_chunk * new_chunk)
{
	new_chunk->next = list_head->next;
	list_head->next = new_chunk;
	list_head->chunk_ct++;
	return;
}


void initialize_tetrahedron(struct tetrahedron * tetra)
{
	int i;
	
	for (i = 0; i < 4; i++) {
		tetra->vertices[i].a = 0.0;
		tetra->vertices[i].b = 0.0;
		tetra->vertices[i].c = 0.0;
	}
	for (i = 0; i < 4; i++)
		tetra->vertex_order[i] = i;	// top-second-third-bottom
	tetra->locator.a = 0;
	tetra->locator.b = 0;
	tetra->locator.c = 0;
	tetra->height = 0.0;
	tetra->gradient = HNV_UNKNOWN_G;		// xyz, xyw, xzw, yzw
	tetra->belongs_to = HNV_UNKNOWN_C;		// class: IN, BD, OUT, UNKNOWN
	tetra->direction = 0;			// UP (= +1) or DOWN (= -1)
	tetra->nearby_peak = -1;
	
	tetra->tetra_ct = 0;		// count of tetrahedrons (for list_head)
        tetra->next = NULL;
	return;
}


//
// tetra_pool_head -> new_chunk    -> old_head_chunk ->
//                      |
//                    tetra_array[initial_tetra_chunk_size]
//
int allocate_tetrahedrons(void)
{
	struct tetrahedron_chunk * new_chunk;
	struct tetrahedron * new_tetra_array;
	int t_ct;
	
	if (get_param_safe_limit(&my_param) < get_param_init_size(&my_param) + 
	      tetrahedron_pool_head.chunk_ct * get_param_incre_size(&my_param)) {
		fprintf(STDERR, 
			"allocate_tetrahedrons() : ERROR  num of tetrahedrons is over the safe limit (= %d)\n", 
			get_param_safe_limit(&my_param));
		return FAIL;
	}
	new_chunk = (struct tetrahedron_chunk *) malloc(sizeof(struct tetrahedron_chunk));
	if (new_chunk == NULL) {
		fprintf(STDERR, 
			"allocate_tetrahedrons() : ERROR allocation failed (new_chunk)\n");
		return FAIL;
	}
	initialize_tetra_chunk(new_chunk);
	if (tetrahedron_pool_head.next == NULL) {
		new_tetra_array = (struct tetrahedron *) malloc(get_param_init_size(&my_param) 
					 * sizeof(struct tetrahedron));
		if (new_tetra_array == NULL) {
			fprintf(STDERR, 
				"allocate_tetrahedrons() : ERROR initial allocation failed (new_tetra_array[%d])\n",
				get_param_init_size(&my_param));
			return FAIL;
		}
		new_chunk->array_size = get_param_init_size(&my_param);
	} else {
		new_tetra_array = (struct tetrahedron *) malloc(get_param_incre_size(&my_param) 
					 * sizeof(struct tetrahedron));
		if (new_tetra_array == NULL) {
			fprintf(STDERR, 
				"allocate_tetrahedrons() : ERROR incremental allocation failed (new_tetra_array[%d])\n",
				get_param_incre_size(&my_param));
			return FAIL;
		}
		new_chunk->array_size = get_param_incre_size(&my_param);
	}
	for (t_ct = 0; t_ct < new_chunk->array_size; t_ct++)
		initialize_tetrahedron(new_tetra_array + t_ct);
	new_chunk->tetra_array = new_tetra_array;
	put_tetra_chunk(&tetrahedron_pool_head, new_chunk);
	return SUCCESS;
}


void copy_vertex(struct hnv_coor_3d * tgt_vertex, struct hnv_coor_3d * src_vertex)
{
	tgt_vertex->a = src_vertex->a;
	tgt_vertex->b = src_vertex->b;
	tgt_vertex->c = src_vertex->c;
	return;
}


void print_tetra_chunk(struct tetrahedron_chunk * t_chunk)
{
	fprintf(STDOUT, "print_tetra_chunk() : chunk = %x, array_size = %d, chunk_ct = %d, next chunk = %x\n",
		(unsigned int) t_chunk, t_chunk->array_size, t_chunk->chunk_ct, (unsigned int) t_chunk->next);
	return;
}


void print_tetra_chunk_list(struct tetrahedron_chunk * chunk)
{
	struct tetrahedron_chunk * next_chunk;
	int c_ct;
	
	next_chunk = chunk;
	c_ct = 0;
	while (next_chunk != NULL) {
		fprintf(STDOUT, "print_tetra_chunk_list() : start of %d-th tetra chunk\n", c_ct);
		print_tetra_chunk(next_chunk);
		next_chunk = next_chunk->next;
		c_ct++;
	}
	return;
}


//===============================================
// Public methos
//===============================================

void print_tetrahedron(struct tetrahedron * ft)
{
	int i;
	
	for (i = 0; i < 4; i++) {
		fprintf(STDOUT, "print_tetrahedron() : vector[%d] = { %8.3f, %8.3f, %8.3f }\n",
			i, ft->vertices[i].a, ft->vertices[i].b, ft->vertices[i].c);
	}
	fprintf(STDOUT, "print_tetrahedron() : vector_order =  %d -> %d -> %d -> %d\n",
		ft->vertex_order[0], ft->vertex_order[1], ft->vertex_order[2], ft->vertex_order[3]);
	fprintf(STDOUT, "print_tetrahedron() : locator = { %d, %d, %d }, tetra = %x\n",
		ft->locator.a, ft->locator.b, ft->locator.c, (unsigned int) ft);
	fprintf(STDOUT, "print_tetrahedron() : height = %8.3f,", ft->height);
	switch (ft->gradient) {
		case HNV_XYZ:
			fprintf(STDOUT, "  gradient = XYZ,");
			break;
		case HNV_XYW:
			fprintf(STDOUT, "  gradient = XYW,");
			break;
		case HNV_XZW:
			fprintf(STDOUT, "  gradient = XZW,");
			break;
		case HNV_YZW:
			fprintf(STDOUT, "  gradient = YZW,");
			break;
		case HNV_UNKNOWN_G:
			fprintf(STDOUT, "  gradient = UNKNOWN,");
			break;
		default:
			fprintf(STDOUT, "\nprint_tetrahedron() : ERROR gradient = invalid value\n");
			break;
	}
	switch (ft->belongs_to) {
		case HNV_IN:
			fprintf(STDOUT, "  belongs_to = IN,");
			break;
		case HNV_OUT:
			fprintf(STDOUT, "  belongs_to = OUT,");
			break;
		case HNV_BD:
			fprintf(STDOUT, "  belongs_to = BD,");
			break;
		case HNV_UNKNOWN_C:
			fprintf(STDOUT, "  belongs_to = UNKNOWN,");
			break;
		default:
			fprintf(STDOUT, "\nprint_tetrahedron() : ERROR belongs_to = invalid value\n");
			break;
	}
	switch (ft->direction) {
		case HNV_UP:
			fprintf(STDOUT, "  direction = UP\n");
			break;
		case HNV_DOWN:
			fprintf(STDOUT, "  direction = DOWN\n");
			break;
		case 0:
			fprintf(STDOUT, "  direction = UNKNOWN\n");
			break;
		default:
			fprintf(STDOUT, "\nprint_tetrahedron() : ERROR direction = invalid value\n");
			break;
	}
	fprintf(STDOUT, "print_tetrahedron() : nearby_peak = %d,", ft->nearby_peak);
	
	fprintf(STDOUT, "  tetra_ct = %d,", ft->tetra_ct);
	fprintf(STDOUT, "  ft->next = %x\n", (unsigned int) ft->next);
        return;
}


int get_tetra_tetra_ct(struct tetrahedron * ft)
{
	return ft->tetra_ct;
}


void print_tetra_list(struct tetrahedron * tetra)
{
	struct tetrahedron * next_tetra;
	int t_ct;
	int tetra_ct;
	
	tetra_ct = get_tetra_tetra_ct(tetra);
	fprintf(STDOUT, "print_tetra_list() : length = %d\n", tetra->tetra_ct);
	next_tetra = tetra->next;	// skip the first tetra
	t_ct = 0;
	while (next_tetra != NULL) {
		if (t_ct > tetra_ct) {
			fprintf(STDERR, "print_tetra_list() : ERROR overflow  t_ct (=%d) >= list length (=%d), current tetra = %x\n",
				t_ct, tetra_ct, (unsigned int) next_tetra);
			exit(1);
		}
		fprintf(STDOUT, "print_tetra_list() : start of %d-th tetrahedron\n", t_ct);
		print_tetrahedron(next_tetra);
		next_tetra = next_tetra->next;
		t_ct++;
	}
	return;
}


HNV_TETRA_CLASS get_tetra_class(struct tetrahedron * ft)
{
	return ft->belongs_to;
}


void set_tetra_class(struct tetrahedron * ft, HNV_TETRA_CLASS value)
{
	ft->belongs_to = value;
	return;
}


void set_tetra_vertex(struct tetrahedron * ft, struct hnv_coor_3d * top, 
		      struct hnv_coor_3d * second, struct hnv_coor_3d * third, 
		      struct hnv_coor_3d * bottom)
{
	ft->vertices[0].a = top->a;
	ft->vertices[0].b = top->b;
	ft->vertices[0].c = top->c;
	ft->vertices[1].a = second->a;
	ft->vertices[1].b = second->b;
	ft->vertices[1].c = second->c;
	ft->vertices[2].a = third->a;
	ft->vertices[2].b = third->b;
	ft->vertices[2].c = third->c;
	ft->vertices[3].a = bottom->a;
	ft->vertices[3].b = bottom->b;
	ft->vertices[3].c = bottom->c;
	return;
}


void get_tetra_vertex(struct hnv_coor_3d * second, struct tetrahedron * ft, int ind)
{
	second->a = ft->vertices[ind].a;
	second->b = ft->vertices[ind].b;
	second->c = ft->vertices[ind].c;
	return;
}


int get_tetra_top_vertex(struct hnv_coor_3d * top, struct tetrahedron * ft)
{
	top->a = ft->vertices[ft->vertex_order[0]].a;
	top->b = ft->vertices[ft->vertex_order[0]].b;
	top->c = ft->vertices[ft->vertex_order[0]].c;
	return ft->vertex_order[0];
}


void get_tetra_second_vertex(struct hnv_coor_3d * second, struct tetrahedron * ft)
{
	second->a = ft->vertices[ft->vertex_order[1]].a;
	second->b = ft->vertices[ft->vertex_order[1]].b;
	second->c = ft->vertices[ft->vertex_order[1]].c;
	return;
}

void get_tetra_third_vertex(struct hnv_coor_3d * third, struct tetrahedron * ft)
{
	third->a = ft->vertices[ft->vertex_order[2]].a;
	third->b = ft->vertices[ft->vertex_order[2]].b;
	third->c = ft->vertices[ft->vertex_order[2]].c;
	return;
}

int get_tetra_bottom_vertex(struct hnv_coor_3d * bottom, struct tetrahedron * ft)
{
	bottom->a = ft->vertices[ft->vertex_order[3]].a;
	bottom->b = ft->vertices[ft->vertex_order[3]].b;
	bottom->c = ft->vertices[ft->vertex_order[3]].c;
	return ft->vertex_order[3];
}


// ft-> locator == 4 * center of the tetrahedron
void set_tetra_locator(struct tetrahedron * ft)
{
	ft->locator.a = (int) (ft->vertices[0].a +ft->vertices[1].a +ft->vertices[2].a +ft->vertices[3].a);
	ft->locator.b = (int) (ft->vertices[0].b +ft->vertices[1].b +ft->vertices[2].b +ft->vertices[3].b);
	ft->locator.c = (int) (ft->vertices[0].c +ft->vertices[1].c +ft->vertices[2].c +ft->vertices[3].c);
	return;
}


void get_tetra_locator(struct hnv_coor_3d_int * locator, struct tetrahedron * ft)
{
	locator->a = ft->locator.a;
	locator->b = ft->locator.b;
	locator->c = ft->locator.c;
	return;
}


void set_tetra_direction(struct tetrahedron * ft, int value)
{
	ft->direction = value;
	return;
}


int get_tetra_direction(struct tetrahedron * ft)
{
	return ft->direction;
}


HNV_GRADIENT get_tetra_gradient(struct tetrahedron * ft)
{
	return ft->gradient;
}


void set_tetra_gradient(struct tetrahedron * ft, HNV_GRADIENT value)
{
	ft->gradient = value;
	return;
}


void set_tetra_nearby_peak(struct tetrahedron * ft, int peak_ct)
{
	ft->nearby_peak = peak_ct;
	return;
}


int get_tetra_nearby_peak(struct tetrahedron * ft)
{
	return ft->nearby_peak;
}


void set_tetra_next_tetra(struct tetrahedron * ft, struct tetrahedron * ft_new)
{
	ft->next = ft_new;
	return;
}


struct tetrahedron * get_tetra_next_tetra(struct tetrahedron * ft)
{
	return ft->next;
}


void set_tetra_height(struct tetrahedron * ft, float height)
{
	ft->height = height;
	return;
}


float get_tetra_height(struct tetrahedron * ft)
{
	return ft->height;
}


void copy_tetrahedron(struct tetrahedron * tgt_ft, struct tetrahedron * src_ft)
{
	int i;
	
	for (i = 0; i < 4; i++) {
		tgt_ft->vertices[i].a = src_ft->vertices[i].a;
		tgt_ft->vertices[i].b = src_ft->vertices[i].b;
		tgt_ft->vertices[i].c = src_ft->vertices[i].c;
	}
	for (i = 0; i < 4; i++) {
		tgt_ft->vertex_order[i] = src_ft->vertex_order[i];
	}
	tgt_ft->locator.a = src_ft->locator.a;
	tgt_ft->locator.b = src_ft->locator.b;
	tgt_ft->locator.c = src_ft->locator.c;
	tgt_ft->height = src_ft->height;
	tgt_ft->gradient = src_ft->gradient;
	tgt_ft->belongs_to = src_ft->belongs_to;
	tgt_ft->direction = src_ft->direction;
	tgt_ft->nearby_peak = src_ft->nearby_peak;
	tgt_ft->tetra_ct = src_ft->tetra_ct;
	tgt_ft->next = src_ft->next;
	return;
}


// ft->vertex_order[0]= vertex of smallest hi;
// ft->vertex_order[3]= vertex of largest hi;
void order_tetra_vertices(struct tetrahedron * ft, float * height_std)
{
	int index[4] ={0, 1, 2, 3};
	float min;
	int left, right, save_left;
	int cur_min_index;
	int i;
	
	for (left = 0; left < 3; left++) {
		min = height_std[index[left]];
		cur_min_index = left;
		for (right = left + 1; right < 4; right++) {
			if (min > height_std[index[right]]) {
				min = height_std[index[right]];
				cur_min_index = right;
			}
		}
		save_left = index[left];
		index[left] = index[cur_min_index];
		index[cur_min_index] = save_left;
	}
	for (i = 0; i < 4; i++)
		ft->vertex_order[i] = index[i];
	return;
}


//
// list_head -> NULL
//
void set_up_empty_tetra_list(struct tetrahedron * list_head)
{
	initialize_tetrahedron(list_head);
	return;
}


void put_tetrahedron(struct tetrahedron * list_head, struct tetrahedron * new_ft)
{
	new_ft->next = list_head->next;
	list_head->next = new_ft;
	list_head->tetra_ct++;
	return;
}


//
// list_head -> ... -> (top_chunk->tetra1) -> (top_chunk->tetra0) -> old_tetras
//
void fill_tetra_list(struct tetrahedron * list_head)
{
	struct tetrahedron_chunk * top_chunk;
	int t_ct;
	
	top_chunk = tetrahedron_pool_head.next;
	for (t_ct=0; t_ct < top_chunk->array_size ; t_ct++)
		put_tetrahedron(list_head, top_chunk->tetra_array + t_ct);
	return;
}


void initialize_tetra_pool(void)
{
	initialize_tetra_chunk(&tetrahedron_pool_head);
	if (allocate_tetrahedrons() == FAIL) {
		fprintf(STDERR, "initialize_tetra_pool() : ERROR allocation failed\n");
		exit(1);
	}
	set_up_empty_tetra_list(&unused_terahedron_head);
	fill_tetra_list(&unused_terahedron_head);
	set_up_empty_tetra_list(&bd_tetra_list_head);
	return;
}


void finalize_tetra_pool(void)
{
	struct tetrahedron_chunk * next_chunk;
	struct tetrahedron_chunk * save_chunk;
	
	if (0) {
		fprintf(STDOUT, "finalize_tetra_pool() : DEBUG START of tetrahedron_pool_head\n");
		print_tetra_chunk_list(&tetrahedron_pool_head);
		fprintf(STDOUT, "finalize_tetra_pool() : DEBUG END of tetrahedron_pool_head\n");
		fprintf(STDOUT, "\n");
	}
	next_chunk = tetrahedron_pool_head.next;
	while (next_chunk != NULL) {
		save_chunk = next_chunk;
		next_chunk = next_chunk->next;
		free(save_chunk);
	}
	set_up_empty_tetra_list(&unused_terahedron_head);
	set_up_empty_tetra_list(&bd_tetra_list_head);
	return;
}


struct tetrahedron * take_tetrahedron(struct tetrahedron * list_head)
{
	struct tetrahedron * ft;
	
	ft = list_head->next;
	if (ft == NULL)
		return NULL;
	list_head->next = ft->next;
	ft->next = NULL;
	list_head->tetra_ct--;
	return ft;
}


struct tetrahedron * assign_tetrahedron(void)
{
	struct tetrahedron * ft;
	
	ft = take_tetrahedron(&unused_terahedron_head);
	if (ft == NULL) {
		if (DEBUG)
			fprintf(STDOUT, "assign_tetrahedron() : DEBUG allocate new chunk\n");
		if (allocate_tetrahedrons() == FAIL) {
			fprintf(STDERR, "assign_tetrahedron() : ERROR allocation failed\n");
			exit(1);
		}
		fill_tetra_list(&unused_terahedron_head);
		ft = take_tetrahedron(&unused_terahedron_head);
		if (ft == NULL) {
			fprintf(STDERR, "assign_tetrahedron() : ERROR tetra == NULL\n");
			exit(1);
		}
	}
	return ft;
}


void recall_tetrahedron(struct tetrahedron * ft)
{
	initialize_tetrahedron(ft);
	put_tetrahedron(&unused_terahedron_head, ft);
	return;
}



void compute_next_tetrahedron(struct tetrahedron * ft_tmp, struct tetrahedron * ft)
{
	if (get_tetra_direction(ft) == HNV_UP) {	// go upward
					// q0->q1->q2->q3  ==> p->q0->q1->q2 or q0->q1->q2->p
					// |p| = mirror image of |q3| w.r.t. plane q0q1q2
		copy_vertex(ft_tmp->vertices + 0, ft->vertices + ft->vertex_order[0]);
		copy_vertex(ft_tmp->vertices + 1, ft->vertices + ft->vertex_order[1]);
		copy_vertex(ft_tmp->vertices + 2, ft->vertices + ft->vertex_order[2]);
		ft_tmp->vertices[3].a = ft->vertices[ft->vertex_order[0]].a 
					+ ft->vertices[ft->vertex_order[2]].a 
					- ft->vertices[ft->vertex_order[3]].a;
		ft_tmp->vertices[3].b = ft->vertices[ft->vertex_order[0]].b 
					+ ft->vertices[ft->vertex_order[2]].b 
					- ft->vertices[ft->vertex_order[3]].b;
		ft_tmp->vertices[3].c = ft->vertices[ft->vertex_order[0]].c 
					+ ft->vertices[ft->vertex_order[2]].c 
					- ft->vertices[ft->vertex_order[3]].c;
	} else {	// go downward
			// q0->q1->q2->q3  ==> p->q1->q2->q3 or q1->q2->q3->p
			// |p| = mirror image of |q0| w.r.t. plane q1q2q3
		ft_tmp->vertices[0].a = ft->vertices[ft->vertex_order[3]].a 
					+ ft->vertices[ft->vertex_order[1]].a 
					- ft->vertices[ft->vertex_order[0]].a;
		ft_tmp->vertices[0].b = ft->vertices[ft->vertex_order[3]].b 
					+ ft->vertices[ft->vertex_order[1]].b 
					- ft->vertices[ft->vertex_order[0]].b;
		ft_tmp->vertices[0].c = ft->vertices[ft->vertex_order[3]].c 
					+ ft->vertices[ft->vertex_order[1]].c 
					- ft->vertices[ft->vertex_order[0]].c;
		copy_vertex(ft_tmp->vertices + 1, ft->vertices + ft->vertex_order[1]);
		copy_vertex(ft_tmp->vertices + 2, ft->vertices + ft->vertex_order[2]);
		copy_vertex(ft_tmp->vertices + 3, ft->vertices + ft->vertex_order[3]);
	}
	set_tetra_locator(ft_tmp);
	return;
}


void compute_previous_tetrahedron(struct tetrahedron * ft_tmp, 
				  struct tetrahedron * ft)
{
	if (get_tetra_direction(ft) == HNV_UP) {	// go downward
							// q0->q1->q2->q3  ==> p->q1->q2->q3 or q1->q2->q3->p
							// |p| = mirror image of |q0| w.r.t. plane q1q2q3
		ft_tmp->vertices[0].a = ft->vertices[ft->vertex_order[3]].a 
					+ ft->vertices[ft->vertex_order[1]].a 
					- ft->vertices[ft->vertex_order[0]].a;
		ft_tmp->vertices[0].b = ft->vertices[ft->vertex_order[3]].b 
					+ ft->vertices[ft->vertex_order[1]].b 
					- ft->vertices[ft->vertex_order[0]].b;
		ft_tmp->vertices[0].c = ft->vertices[ft->vertex_order[3]].c 
					+ ft->vertices[ft->vertex_order[1]].c 
					- ft->vertices[ft->vertex_order[0]].c;
		copy_vertex(ft_tmp->vertices + 1, ft->vertices + ft->vertex_order[1]);
		copy_vertex(ft_tmp->vertices + 2, ft->vertices + ft->vertex_order[2]);
		copy_vertex(ft_tmp->vertices + 3, ft->vertices + ft->vertex_order[3]);
	} else {	// go upward
			// q0->q1->q2->q3  ==> p->q0->q1->q2 or q0->q1->q2->p
			// |p| = mirror image of |q3| w.r.t. plane q0q1q2
		copy_vertex(ft_tmp->vertices + 0, ft->vertices + ft->vertex_order[0]);
		copy_vertex(ft_tmp->vertices + 1, ft->vertices + ft->vertex_order[1]);
		copy_vertex(ft_tmp->vertices + 2, ft->vertices + ft->vertex_order[2]);
		ft_tmp->vertices[3].a = ft->vertices[ft->vertex_order[0]].a 
					+ ft->vertices[ft->vertex_order[2]].a 
					- ft->vertices[ft->vertex_order[3]].a;
		ft_tmp->vertices[3].b = ft->vertices[ft->vertex_order[0]].b 
					+ ft->vertices[ft->vertex_order[2]].b 
					- ft->vertices[ft->vertex_order[3]].b;
		ft_tmp->vertices[3].c = ft->vertices[ft->vertex_order[0]].c 
					+ ft->vertices[ft->vertex_order[2]].c
					- ft->vertices[ft->vertex_order[3]].c;
	}
	set_tetra_locator(ft_tmp);
	return;
}